#include "polyline.h"

PolyLine::PolyLine()
{
}

void PolyLine::Add(const QVector2D & p)
{
    m_line.push_back(p);
}

int PolyLine::Size() const
{
    return m_line.size();
}

void PolyLine::Clear()
{
    m_line.clear();
}

const QVector2D & PolyLine::operator[](const int ind) const
{
    return m_line[ind];
}

QVector2D & PolyLine::operator[](const int ind)
{
    return m_line[ind];
}

void PolyLine::DrawGL() const
{

    glPointSize(5.0f);
    glBegin(GL_POINTS);

    for (int i=0; i<m_line.size(); ++i)
        glVertex2f(m_line[i].x(), m_line[i].y());

    glEnd();
    glPointSize(1.0f);

    glBegin(GL_LINE_STRIP);

    for (int i=0; i<m_line.size(); ++i)
        glVertex2f(m_line[i].x(), m_line[i].y());

    glEnd();
}

void PolyLine::Scale(const float s)
{
    for (int i=0; i<m_line.size(); ++i)
        m_line[i] = m_line[i] * s;
}

void PolyLine::Smooth(const int sWin)
{

    QVector <QVector2D> oldline = m_line;

    m_line.clear();

    //ensure first endpoint is the same
    //m_line.push_back(oldline.first());

    for (int i=0; i<oldline.size(); ++i) {

        QVector2D avgPt(0.0f, 0.0f);
        float avgTot = 0.0f;

        //add left to average
        for (int j=i; j>=0; --j) {
            const float dist = (oldline[i] - oldline[j]).length();

            if (dist < float(sWin)) {
                //const float wt = 1.0f / (1.0f + dist);
                const float wt = 1.0f;
                avgPt += oldline[j] * wt;
                avgTot += wt;
            }
            else
                break;

        }

        //add right to average
        int j;
        for (j=i+1; j<oldline.size(); ++j) {

            const float dist = (oldline[i] - oldline[j]).length();

            if (dist < float(sWin)) {
                //const float wt = 1.0f / (1.0f + dist);
                const float wt = 1.0f;
                avgPt += oldline[j] * wt;
                avgTot += wt;
            }
            else
                break;

        }

        //process vertex outside neighbourhood next
        i = j-1;

        m_line.push_back(avgPt / avgTot);

    }

    //ensure last endpoint is the same
    m_line.push_back(oldline.last());

}

void PolyLine::ArcLengths(QVector <float> & arclen) const
{

    arclen.clear();

    float len = 0.0f;
    arclen.push_back(0.0f);

    for (int i=1; i<m_line.size(); ++i) {
        len += (m_line[i] - m_line[i-1]).length();
        arclen.push_back(len);
    }

}

void PolyLine::Curvatures(QVector <float> & curv) const
{

    curv.clear();

    curv.push_back(0.0f);

    for (int i=1; i<m_line.size()-1; ++i) {

        float val = PointCurvature(m_line[i-1], m_line[i], m_line[i+1]);
        curv.push_back(val);

    }

    curv.push_back(0.0f);

    //special case the endpoint curvatures
    //if curve has only 2 points, curvature is 0 (a line segment)
    //if curve has only 3 points, curvature is same at endpoints as interior point (circular arc)
    //else, we linterp the curvature value
    switch (m_line.size()) {

    case 2:
        //do nothing
        break;

    case 3:
        //make circle arc
        curv[0] = curv[1];
        curv[2] = curv[1];
        break;

    default:

        //linterp curvature endpoints using interior neighbour's slope
        float l1 = (m_line[0] - m_line[1]).length();
        float l2 = (m_line[2] - m_line[1]).length();
        float m = (curv[2] - curv[1]) / l2;
        curv[0] = curv[1] - l1 * m;

        const float sz = m_line.size();
        l1 = (m_line[sz - 3] - m_line[sz - 2]).length();
        l2 = (m_line[sz - 2] - m_line[sz - 1]).length();
        m = (curv[sz - 2] - curv[sz - 3])/ l1;
        curv[sz - 1] = curv[sz - 2] + l2 * m;

        break;

    }

}

float PolyLine::PointCurvature(const QVector2D p1, const QVector2D p2, const QVector2D p3)
{

    //this curvature definition from schneider/kobbelt paper
    const QVector2D v1 = p2 - p1;
    const QVector2D v2 = p3 - p2;
    const float l1 = v1.length();
    const float l2 = v2.length();
    const float l12 = (v1 + v2).length();

    float det = v1.x() * v2.y() - v2.x() * v1.y();
    float val = 2.0f * det / (l1 * l2 * l12);

    return val;


}

QVector2D PolyLine::StartTangent() const
{
    if (m_line.size() >= 2) {
        QVector2D tan = m_line[1] - m_line[0];
        return tan.normalized();
    }
    else
        return QVector2D(1,0);
}

QVector2D PolyLine::Tangent(const int ind) const
{
    if (m_line.size() > ind+1) {
        QVector2D tan = m_line[ind+1] - m_line[ind];
        return tan.normalized();
    }
    else
        return QVector2D(1,0);
}

void PolyLine::LoadContours(const QString & path, QVector <QVector <PolyLine> > & contours)
{

    QFile file(path);

    //open file
    if (!file.open( QIODevice::ReadOnly | QIODevice::Text )) {
        qDebug() << "Contour File " << path << " doesn't exist.";
        return;
    }

    QTextStream ifs(&file);

    //read # cuts
    const int nPlanes = ifs.readLine().toInt();

    for (int i=0; i<nPlanes; ++i) {

        //read plane cut params
        const QStringList pparams = ifs.readLine().split(" ");
        const QVector3D pnorm(pparams[0].toFloat(),
                              pparams[1].toFloat(),
                              pparams[2].toFloat());

        //compute basis vectors
        QVector3D basis1;
        QVector3D basis2;

        if (QVector3D::crossProduct(pnorm, QVector3D(0,1,0)).length() > 0.1)
            basis1 = QVector3D::crossProduct(pnorm, QVector3D(0,1,0)).normalized();
        else
            basis1 = QVector3D::crossProduct(pnorm, QVector3D(1,0,0)).normalized();

        basis2 = QVector3D::crossProduct(pnorm, basis1).normalized();

        //read number of verts/edges
        const QStringList each = ifs.readLine().split(" ");

        const int nVert = each[0].toInt();
        const int nEdge = each[1].toInt();

        //read vertex list
        QVector <QVector2D> verts;
        for (int j=0; j<nVert; ++j) {

            //take each 3d vert and map it to a 2d one...
            const QStringList each = ifs.readLine().split(" ");
            const QVector3D pt3d(each[0].toFloat(),
                                 each[1].toFloat(),
                                 each[2].toFloat());
            const QVector2D pt2d(QVector3D::dotProduct(basis1, pt3d) * 0.02f,
                                 QVector3D::dotProduct(basis2, pt3d) * 0.02f);

            verts.push_back(pt2d);

        }

        //read edge list (and generate one or more polylines)
        QVector <PolyLine> plist;
        PolyLine p;
        int v1 = 0;
        int v2 = 0;
        for (int j=0; j<nEdge; ++j) {

            const QStringList each = ifs.readLine().split(" ");
            v1 = each[0].toInt();
            v2 = each[1].toInt();

            p.Add(verts[v1]);

            if (v2 < v1) {
                p.Add(verts[v2]);
                plist.push_back(p);
                p.Clear();
            }

        }
        if (p.Size() > 1) {
            p.Add(verts[v2]);
            plist.push_back(p);
        }

        contours.push_back(plist);

    }

    //close file
    file.close();

}
